/**
 * Copyright (C) 2006-2007 International Business Machines Corp.
 * Author(s): Alon Bar-Lev <alon.barlev@gmail.com>, based on:
 *            Trevor S. Highland <trevor.highland@gmail.com>
 *            Mike Halcrow <mhalcrow@us.ibm.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <fcntl.h>
#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <openssl/pem.h>
#include <openssl/rsa.h>
#include <openssl/err.h>
#include <openssl/engine.h>
#include <pkcs11-helper-1.0/pkcs11h-certificate.h>
#include <pkcs11-helper-1.0/pkcs11h-openssl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "../include/ecryptfs.h"
#include "../include/decision_graph.h"

struct pkcs11_data {
	char *serialized_id;
	unsigned char *certificate_blob;
	size_t certificate_blob_size;
	char *passphrase;
};

/**
 * ecryptfs_pkcs11_deserialize
 * @pkcs11_data: The deserialized version of the key module data;
 *                internal components pointed to blob memory
 * @blob: The key module-specific state blob
 *
 */
static int ecryptfs_pkcs11_deserialize(struct pkcs11_data *pkcs11_data,
					unsigned char *blob)
{
	size_t serialized_id_length;
	size_t passphrase_length;
	size_t i = 0;
	int rc;

	serialized_id_length = blob[i++] % 256;
	serialized_id_length += blob[i++] << 8;
	pkcs11_data->serialized_id = blob + i;
	i += serialized_id_length;
	pkcs11_data->certificate_blob_size = blob[i++] % 256;
	pkcs11_data->certificate_blob_size += blob[i++] << 8;
	if (pkcs11_data->certificate_blob_size == 0) {
		pkcs11_data->certificate_blob = NULL;
	}
	else {
		pkcs11_data->certificate_blob = blob + i;
		i += pkcs11_data->certificate_blob_size;
	}
	passphrase_length = blob[i++] % 256;
	passphrase_length += blob[i++] << 8;
	pkcs11_data->passphrase = blob + i;
	i += passphrase_length;

	rc = 0;

out:

	return rc;
}

/**
 * @blob: Callee allocates this memory
 */
static int ecryptfs_pkcs11_serialize(unsigned char *blob, size_t *blob_size,
				      struct pkcs11_data *pkcs11_data)
{
	size_t serialized_id_length;
	size_t passphrase_length;
	size_t i = 0;
	int rc = 0;

	(*blob_size) = 0;
	if (!pkcs11_data->serialized_id || !pkcs11_data->passphrase) {
		rc = -EINVAL;
		syslog(LOG_ERR, "pkcs11_data internal structure not "
		       "properly filled in\n");
		goto out;
	}
	serialized_id_length = strlen(pkcs11_data->serialized_id) + 1; /* + '\0' */
	if (blob) blob[i] = serialized_id_length % 256;
	i++;
	if (blob) blob[i] = serialized_id_length >> 8;
	i++;
	if (blob) memcpy(&blob[i], pkcs11_data->serialized_id, serialized_id_length);
	i += serialized_id_length;
	if (blob) blob[i] = pkcs11_data->certificate_blob_size % 256;
	i++;
	if (blob) blob[i] = pkcs11_data->certificate_blob_size >> 8;
	i++;
	if (blob) memcpy(&blob[i], pkcs11_data->certificate_blob, pkcs11_data->certificate_blob_size);
	i += pkcs11_data->certificate_blob_size;
	passphrase_length = strlen(pkcs11_data->passphrase) + 1; /* + '\0' */
	if (blob) blob[i] = passphrase_length % 256;
	i++;
	if (blob) blob[i] = passphrase_length >> 8;
	i++;
	if (blob) memcpy(&blob[i], pkcs11_data->passphrase, passphrase_length);
	i += passphrase_length;
	(*blob_size) = i;
out:
	return rc;
}

static
void
pkcs11_log (
	void * const global_data,
	unsigned flags,
	const char * const format,
	va_list args
) {
	vsyslog(LOG_INFO, format, args);
}

static
PKCS11H_BOOL
pkcs11_token_prompt (
	void * const global_data,
	void * const user_data,
	const pkcs11h_token_id_t token,
	const unsigned retry
) {
	(void)global_data;
	(void)user_data;
	(void)retry;

	return FALSE;
}

static
PKCS11H_BOOL
pkcs11_pin_prompt (
	void * const global_data,
	void * const user_data,
	const pkcs11h_token_id_t token,
	const unsigned retry,
	char * const pin,
	const size_t pin_max
) {
	(void)global_data;

	if (retry == 0 && user_data != NULL) {
		strncpy (pin, (char *)user_data, pin_max-1);
		pin[pin_max-1] = '\x0';

		return TRUE;
	}
	else {
		return FALSE;
	}
}

struct ecryptfs_subgraph_ctx {
	struct ecryptfs_key_mod *key_mod;
	struct pkcs11_data pkcs11_data;
};

static void
ecryptfs_pkcs11_destroy_subgraph_ctx(struct ecryptfs_subgraph_ctx *ctx)
{
	if (ctx->pkcs11_data.serialized_id != NULL) {
		free(ctx->pkcs11_data.serialized_id);
	}
	if (ctx->pkcs11_data.passphrase != NULL) {
		memset(ctx->pkcs11_data.passphrase, 0, strlen(ctx->pkcs11_data.passphrase));
		free(ctx->pkcs11_data.passphrase);
	}
	if (ctx->pkcs11_data.certificate_blob != NULL) {
		free(ctx->pkcs11_data.certificate_blob);
	}
	memset(&ctx->pkcs11_data, 0, sizeof(ctx->pkcs11_data));
	memset(ctx, 0, sizeof(struct ecryptfs_subgraph_ctx));
}

/**
 * ecryptfs_pkcs11_generate_signature
 * @sig: Generated sig (ECRYPTFS_SIG_SIZE_HEX + 1 bytes of allocated memory)
 * @key: RSA key from which to generate sig
 */
static int ecryptfs_pkcs11_generate_signature(char *sig, RSA *key)
{
	int len, nbits, ebits, i;
	int nbytes, ebytes;
	char *hash;
	char *data = NULL;
	int rc = 0;

	hash = malloc(SHA_DIGEST_LENGTH);
	if (!hash) {
		syslog(LOG_ERR, "Out of memory\n");
		rc = -ENOMEM;
		goto out;
	}
	nbits = BN_num_bits(key->n);
	nbytes = nbits / 8;
	if (nbits % 8)
		nbytes++;
	ebits = BN_num_bits(key->e);
	ebytes = ebits / 8;
	if (ebits % 8)
		ebytes++;
	len = 10 + nbytes + ebytes;
	data = malloc(3 + len);
	if (!data) {
		syslog(LOG_ERR, "Out of memory\n");
		rc = -ENOMEM;
		goto out;
	}
	i = 0;
	data[i++] = '\x99';
	data[i++] = (char)(len >> 8);
	data[i++] = (char)len;
	data[i++] = '\x04';
	data[i++] = '\00';
	data[i++] = '\00';
	data[i++] = '\00';
	data[i++] = '\00';
	data[i++] = '\02';
	data[i++] = (char)(nbits >> 8);
	data[i++] = (char)nbits;
	BN_bn2bin(key->n, &(data[i]));
	i += nbytes;
	data[i++] = (char)(ebits >> 8);
	data[i++] = (char)ebits;
	BN_bn2bin(key->e, &(data[i]));
	i += ebytes;
	SHA1(data, len + 3, hash);
	to_hex(sig, hash, ECRYPTFS_SIG_SIZE);
	sig[ECRYPTFS_SIG_SIZE_HEX] = '\0';
out:
	free(data);
	free(hash);
	return rc;
}

/**
 * ecryptfs_pkcs11_read_key
 * @rsa: RSA key to allocate
 * @blob: Key module data to use in finding the key
 */
static int ecryptfs_pkcs11_read_key(RSA **rsa, unsigned char *blob)
{
	struct pkcs11_data _pkcs11_data;
	struct pkcs11_data *pkcs11_data = &_pkcs11_data;
	pkcs11h_certificate_id_t certificate_id = NULL;
	pkcs11h_certificate_t certificate = NULL;
	pkcs11h_openssl_session_t openssl_session = NULL;
	CK_RV rv = CKR_OK;
	int rc;

	if ((rc = ecryptfs_pkcs11_deserialize(pkcs11_data, blob)) != 0) {
		goto out;
	}

	if (
		(rv = pkcs11h_certificate_deserializeCertificateId (
			&certificate_id,
			pkcs11_data->serialized_id
		)) != CKR_OK
	) {
		syslog(LOG_ERR, "PKCS#11: Cannot deserialize id %ld-'%s'", rv, pkcs11h_getMessage (rv));
		rc = -EIO;
		goto out;
	}

	if (
		pkcs11_data->certificate_blob != NULL &&
		(rv = pkcs11h_certificate_setCertificateIdCertificateBlob (
			certificate_id,
			pkcs11_data->certificate_blob,
			pkcs11_data->certificate_blob_size
		)) != CKR_OK
	) {
		syslog(LOG_ERR, "PKCS#11: Cannot set certificate blob %ld-'%s'", rv, pkcs11h_getMessage (rv));
		rc = -EIO;
		goto out;
	}

	if (
		(rv = pkcs11h_certificate_create (
			certificate_id,
			pkcs11_data->passphrase,
			PKCS11H_PROMPT_MASK_ALLOW_ALL,
			PKCS11H_PIN_CACHE_INFINITE,
			&certificate
		)) != CKR_OK
	) {
		syslog(LOG_ERR, "PKCS#11: Cannot get certificate %ld-'%s'", rv, pkcs11h_getMessage (rv));
		rc = -EIO;
		goto out;
	}

	if ((openssl_session = pkcs11h_openssl_createSession (certificate)) == NULL) {
		syslog(LOG_ERR, "PKCS#11: Cannot initialize openssl session");
		rc = -EIO;
		goto out;
	}

	/*
	 * Will be released by openssl_session
	 */
	certificate = NULL;

	if ((*rsa = pkcs11h_openssl_session_getRSA (openssl_session)) == NULL) {
		syslog(LOG_ERR, "PKCS#11: Unable get rsa object");
		rc = -EIO;
		goto out;
	}

	rc = 0;
out:
	if (certificate != NULL) {
		pkcs11h_certificate_freeCertificate (certificate);
		certificate = NULL;
	}

	if (certificate_id != NULL) {
		pkcs11h_certificate_freeCertificateId (certificate_id);
		certificate_id = NULL;
	}
	
	if (openssl_session != NULL) {
		pkcs11h_openssl_freeSession (openssl_session);
		openssl_session = NULL;
	}

	return rc;
}

int ecryptfs_pkcs11_get_key_sig(unsigned char *sig, unsigned char *blob)
{
	RSA *rsa = NULL;
	int rc;

	if ((rc = ecryptfs_pkcs11_read_key(&rsa, blob))) {
		syslog(LOG_ERR, "Error attempting to read RSA key from token;"
		       " rc = [%d]\n", rc);
		goto out;
	}
	if ((rc = ecryptfs_pkcs11_generate_signature(sig, rsa))) {
		syslog(LOG_ERR, "%s: Error attempting to generate key "
		       "signature; rc = [%d]\n", __FUNCTION__, rc);
		goto out_free_rsa;
	}
out_free_rsa:
	RSA_free(rsa);
out:
	return rc;
}

/**
 * ecryptfs_pkcs11_encrypt
 * @to: Where to write encrypted data
 * @size: Number of bytes to encrypt
 * @from: Data to encrypt
 * @blob: Arbitrary blob specific to this key module
 *
 * Encrypt @size bytes of data in @from, writing the encrypted data
 * into @to, using @blob as the parameters for the
 * encryption.
 */
static int ecryptfs_pkcs11_encrypt(char *to, size_t *to_size, char *from,
				    size_t from_size, unsigned char *blob,
				    int blob_type)
{
	RSA *rsa = NULL;
	int rc;

	(*to_size) = 0;
	if ((rc = ecryptfs_pkcs11_read_key(&rsa, blob))) {
		rc = -(int)ERR_get_error();
		syslog(LOG_ERR, "Error attempting to read RSA key from file;"
		       " rc = [%d]\n", rc);
		goto out;
	}
	(*to_size) = RSA_size(rsa);
	if (to) {
		if ((rc = RSA_public_encrypt(from_size, from, to, rsa,
					     RSA_PKCS1_PADDING)) == -1) {
			rc = -(int)ERR_get_error();
			syslog(LOG_ERR, "Error attempting to perform RSA "
			       "public key encryption; rc = [%d]\n", rc);
			goto out_free_rsa;
		} else {
			(*to_size) = rc;
			rc = 0;
		}
	}
out_free_rsa:
	RSA_free(rsa);
out:

	return rc;
}

/**
 * ecryptfs_pkcs11_dencrypt
 * @from: Data to decrypt
 * @to: Where to write decrypted data
 * @decrypted_key_size: Number of bytes decrypted
 * @blob: Arbitrary blob specific to this key module
 *
 * Decrypt data in @from, writing the decrypted data into @to, using
 * @blob as the parameters for the encryption.
 */
static int ecryptfs_pkcs11_decrypt(char *to, size_t *to_size, char *from,
				    size_t from_size, unsigned char *blob,
				    int blob_type)
{
	RSA *rsa = NULL;
	int rc;

	(*to_size) = 0;
	if ((rc = ecryptfs_pkcs11_read_key(&rsa, blob))) {
		rc = -(int)ERR_get_error();
		syslog(LOG_ERR, "Error attempting to read RSA key from file;"
		       " rc = [%d]\n", rc);
		goto out;
	}
	(*to_size) = RSA_size(rsa);
	if (to) {
		if ((rc = RSA_private_decrypt(from_size, from, to, rsa,
					      RSA_PKCS1_PADDING)) == -1) {
			rc = -(int)ERR_get_error();
			syslog(LOG_ERR, "Error attempting to perform RSA "
			       "public key decryption; rc = [%d]\n", rc);
			goto out_free_rsa;
		} else {
			(*to_size) = rc;
			rc = 0;
		}
	}
out_free_rsa:
	RSA_free(rsa);
out:
	return rc;
}

static int
ecryptfs_pkcs11_process_key(struct ecryptfs_subgraph_ctx *subgraph_ctx,
			     struct val_node **mnt_params)
{
	struct pkcs11_data *pkcs11_data = &subgraph_ctx->pkcs11_data;
	pkcs11h_certificate_id_t certificate_id = NULL;
	pkcs11h_certificate_t certificate = NULL;
	size_t blob_size;
	char *sig_mnt_opt;
	char sig[ECRYPTFS_SIG_SIZE_HEX + 1];
	CK_RV rv;
	int rc;

	if (
		(rv = pkcs11h_certificate_deserializeCertificateId (
			&certificate_id,
			pkcs11_data->serialized_id
		)) != CKR_OK
	) {
		syslog(LOG_ERR, "PKCS#11: Cannot deserialize id %ld-'%s'", rv, pkcs11h_getMessage (rv));
		rc = -EIO;
		goto out;
	}

	if (
		(rv = pkcs11h_certificate_create (
			certificate_id,
			pkcs11_data->passphrase,
			PKCS11H_PROMPT_MASK_ALLOW_ALL,
			PKCS11H_PIN_CACHE_INFINITE,
			&certificate
		)) != CKR_OK
	) {
		syslog(LOG_ERR, "PKCS#11: Cannot get certificate %ld-'%s'", rv, pkcs11h_getMessage (rv));
		rc = -EIO;
		goto out;
	}

	if (pkcs11_data->certificate_blob == NULL) {
		if (
			(rv = pkcs11h_certificate_getCertificateBlob (
				certificate,
				NULL,
				&pkcs11_data->certificate_blob_size
			)) != CKR_OK
		) {
			syslog(LOG_ERR, "PKCS#11: Cannot load certificate %ld-'%s'", rv, pkcs11h_getMessage (rv));
			rc = -EIO;
			goto out;
		}

		pkcs11_data->certificate_blob = malloc(pkcs11_data->certificate_blob_size);
		if (!pkcs11_data->certificate_blob) {
			syslog(LOG_ERR, "Out of memory\n");
			rc = -ENOMEM;
			goto out;
		}

		if (
			(rv = pkcs11h_certificate_getCertificateBlob (
				certificate,
				pkcs11_data->certificate_blob,
				&pkcs11_data->certificate_blob_size
			)) != CKR_OK
		) {
			syslog(LOG_ERR, "PKCS#11: Cannot load certificate %ld-'%s'", rv, pkcs11h_getMessage (rv));
			rc = -EIO;
			goto out;
		}
	}

	if ((rc = ecryptfs_pkcs11_serialize(NULL, &blob_size, 
					     pkcs11_data))) {
		syslog(LOG_ERR, "Error serializing pkcs11; rc = [%d]\n", rc);
		rc = MOUNT_ERROR;
		goto out;
	}
	if (blob_size == 0) {
		syslog(LOG_ERR, "Error serializing pkcs11\n");
		rc = MOUNT_ERROR;
		goto out;
	}
	if ((subgraph_ctx->key_mod->blob = malloc(blob_size)) == NULL) {
		syslog(LOG_ERR, "Out of memory\n");
		rc = MOUNT_ERROR;
		goto out;
	}
	if ((rc = ecryptfs_pkcs11_serialize(subgraph_ctx->key_mod->blob,
					     &subgraph_ctx->key_mod->blob_size, 
					     pkcs11_data))) {
		syslog(LOG_ERR, "Error serializing pkcs11; rc = [%d]\n", rc);
		rc = MOUNT_ERROR;
		goto out;
	}
	if (subgraph_ctx->key_mod->blob_size != blob_size) {
		syslog(LOG_ERR, "%s: Internal error\n", __FUNCTION__);
		exit(1);
	}
	if ((rc = ecryptfs_add_key_module_key_to_keyring(
		     sig, subgraph_ctx->key_mod)) < 0) {
		syslog(LOG_ERR, "Error attempting to add key to keyring for "
		       "key module [%s]; rc = [%d]\n",
		       subgraph_ctx->key_mod->alias, rc);
		rc = MOUNT_ERROR;
		goto out;
	}
	if ((rc = asprintf(&sig_mnt_opt, "ecryptfs_sig=%s", sig)) == -1) {
		rc = MOUNT_ERROR;
		goto out;
	}
	rc = 0;
	stack_push(mnt_params, sig_mnt_opt);
out:
	if (certificate != NULL) {
		pkcs11h_certificate_freeCertificate (certificate);
		certificate = NULL;
	}

	if (certificate_id != NULL) {
		pkcs11h_certificate_freeCertificateId (certificate_id);
		certificate_id = NULL;
	}

	return rc;
}

static int tf_pkcs11_id(struct ecryptfs_ctx *ctx, struct param_node *node,
		       struct val_node **mnt_params, void **foo)
{
	struct ecryptfs_subgraph_ctx *subgraph_ctx;
	int rc;

	subgraph_ctx = (struct ecryptfs_subgraph_ctx *)(*foo);
	if ((rc = asprintf(&subgraph_ctx->pkcs11_data.serialized_id, "%s", node->val))
	    == -1) {
		rc = MOUNT_ERROR;
		goto out;
	}
	rc = DEFAULT_TOK;
	node->val = NULL;
out:
	return rc;
}

static int tf_pkcs11_passwd(struct ecryptfs_ctx *ctx, struct param_node *node,
			 struct val_node **mnt_params, void **foo)
{
	struct ecryptfs_subgraph_ctx *subgraph_ctx;
	int rc;

	subgraph_ctx = (struct ecryptfs_subgraph_ctx *)(*foo);
	if ((rc = asprintf(&subgraph_ctx->pkcs11_data.passphrase, "%s",
			   node->val)) == -1) {
		rc = MOUNT_ERROR;
		goto out;
	}
	node->val = NULL;
	rc = DEFAULT_TOK;
out:
	return rc;
}

static int tf_pkcs11_x509file(struct ecryptfs_ctx *ctx, struct param_node *node,
			 struct val_node **mnt_params, void **foo)
{
	struct ecryptfs_subgraph_ctx *subgraph_ctx;
	X509 *x509 = NULL;
	unsigned char *p = NULL;
	FILE *fp = NULL;
	int rc;

	subgraph_ctx = (struct ecryptfs_subgraph_ctx *)(*foo);

	if (node->val != NULL && strlen (node->val) > 0) {
		if ((fp = fopen (node->val, "r")) == NULL) {
			syslog (LOG_ERR, "PKCS#11: Cannot open file '%s'", node->val);
			rc = -errno;
			goto out;
		}

		if (
			!PEM_read_X509 (
				fp,
				&x509,
				NULL,
				0
			)
		) {
			x509 = NULL;
			syslog (LOG_ERR, "PKCS#11: Cannot read PEM from file '%s'", node->val);
			rc = -EIO;
			goto out;
		}

		if ((subgraph_ctx->pkcs11_data.certificate_blob_size = i2d_X509 (x509, NULL)) < 0	) {
			syslog (LOG_ERR, "PKCS#11: Cannot read decode certificate");
			rc = -EIO;
			goto out;
		}

		if ((subgraph_ctx->pkcs11_data.certificate_blob = (unsigned char *)malloc (subgraph_ctx->pkcs11_data.certificate_blob_size)) == NULL) {
			syslog (LOG_ERR, "PKCS#11: Cannot allocate memory");
			rc = -ENOMEM;
			goto out;
		}

		/*
		 * i2d_X509 increments p!!!
		 */
		p = subgraph_ctx->pkcs11_data.certificate_blob;

		if ((subgraph_ctx->pkcs11_data.certificate_blob_size = i2d_X509 (x509, &p)) < 0) {
			error ("PKCS#11: Cannot read decode certificate");
			goto out;
		}
	}

	node->val = NULL;
	if ((rc = ecryptfs_pkcs11_process_key(subgraph_ctx, mnt_params))) {
		syslog(LOG_ERR, "Error processing PKCS#11 key; rc = [%d]", rc);
		goto out;
	}
	ecryptfs_pkcs11_destroy_subgraph_ctx(subgraph_ctx);
	free(subgraph_ctx);
	(*foo) = NULL;
	rc = DEFAULT_TOK;

out:

	if (x509 != NULL) {
		X509_free (x509);
		x509 = NULL;
	}

	if (fp != NULL) {
		fclose (fp);
		fp = NULL;
	}

	return rc;
}

#define PKCS11_TOK 0
#define PKCS11_ID_TOK 1
#define PKCS11_PASSWD_TOK 2
#define PKCS11_PASS_ENV_TOK 3
#define PKCS11_PASS_STDIN_TOK 4
#define PKCS11_DEFAULT_PASS_TOK 5
#define PKCS11_DEFAULT_X509_FILE 6
static struct param_node pkcs11_param_nodes[] = {
	{.num_mnt_opt_names = 1,
	 .mnt_opt_names = {"keyformat"},
	 .prompt = "Key format",
	 .val_type = VAL_STR,
	 .val = NULL,
	 .display_opts = NULL,
	 .default_val = "id",
	 .flags = ECRYPTFS_PARAM_FLAG_NO_VALUE,
	 .num_transitions = 1,
	 .tl = {{.val = "default",
		 .pretty_val = "PKCS#11 ID",
		 .next_token = &pkcs11_param_nodes[PKCS11_ID_TOK],
		 .trans_func = NULL}}},

	{.num_mnt_opt_names = 1,
	 .mnt_opt_names = {"id"},
	 .prompt = "PKCS#11 Serialized ID",
	 .val_type = VAL_STR,
	 .val = NULL,
	 .display_opts = NULL,
	 .default_val = NULL,
	 .suggested_val = NULL,
	 .flags = ECRYPTFS_PARAM_FLAG_ECHO_INPUT,
	 .num_transitions = 4,
	 .tl = {{.val = "passwd",
		 .pretty_val = "",
		 .next_token = &pkcs11_param_nodes[PKCS11_PASSWD_TOK],
		 .trans_func = tf_pkcs11_id},
	 	{.val = "passenv",
		 .pretty_val = "Passphrase ENV",
		 .next_token = &pkcs11_param_nodes[PKCS11_PASS_ENV_TOK],
		 .trans_func = tf_pkcs11_id},
	 	{.val = "passstdin",
		 .pretty_val = "Passphrase STDIN",
		 .next_token = &pkcs11_param_nodes[PKCS11_PASS_STDIN_TOK],
		 .trans_func = tf_pkcs11_id},
	 	{.val = "default",
		 .pretty_val = "Passphrase",
		 .next_token = &pkcs11_param_nodes[PKCS11_DEFAULT_PASS_TOK],
		 .trans_func = tf_pkcs11_id}}},

	{.num_mnt_opt_names = 1,
	 .mnt_opt_names = {"passwd"},
	 .prompt = "Passphrase",
	 .val_type = VAL_STR,
	 .val = NULL,
	 .display_opts = NULL,
	 .default_val = NULL,
	 .flags = ECRYPTFS_PARAM_FLAG_MASK_OUTPUT,
	 .num_transitions = 1,
	 .tl = {{.val = NULL,
		 .pretty_val = NULL,
		 .next_token = NULL,
		 .trans_func = tf_pkcs11_passwd}}},

	{.num_mnt_opt_names = 1,
	 .mnt_opt_names = {"passenv"},
	 .prompt = "Passphrase",
	 .val_type = VAL_STR,
	 .val = NULL,
	 .display_opts = NULL,
	 .default_val = NULL,
	 .flags = ECRYPTFS_PARAM_FLAG_MASK_OUTPUT,
	 .num_transitions = 1,
	 .tl = {{.val = NULL,
		 .pretty_val = NULL,
		 .next_token = NULL,
		 .trans_func = tf_pkcs11_passwd}}},

	{.num_mnt_opt_names = 1,
	 .mnt_opt_names = {"passstdin"},
	 .prompt = "Passphrase",
	 .val_type = VAL_STR,
	 .val = NULL,
	 .display_opts = NULL,
	 .default_val = NULL,
	 .flags = VERIFY_VALUE | STDIN_REQUIRED,
	 .num_transitions = 1,
	 .tl = {{.val = NULL,
		 .pretty_val = NULL,
		 .next_token = NULL,
		 .trans_func = tf_pkcs11_passwd}}},

	{.num_mnt_opt_names = 1,
	 .mnt_opt_names = {"defaultpass"},
	 .prompt = "Passphrase",
	 .val_type = VAL_STR,
	 .val = NULL,
	 .display_opts = NULL,
	 .default_val = NULL,
	 .flags = STDIN_REQUIRED,
	 .num_transitions = 1,
	 .tl = {{.val = "default",
		 .pretty_val = "Optional X.509 Certificate PEM file",
		 .next_token = &pkcs11_param_nodes[PKCS11_DEFAULT_X509_FILE],
		 .trans_func = tf_pkcs11_passwd}}},

	{.num_mnt_opt_names = 1,
	 .mnt_opt_names = {"x509file"},
	 .prompt = "Optional X.509 Certificate PEM file",
	 .val_type = VAL_STR,
	 .val = NULL,
	 .display_opts = NULL,
	 .default_val = NULL,
	 .flags = ECRYPTFS_PARAM_FLAG_ECHO_INPUT,
	 .num_transitions = 1,
	 .tl = {{.val = NULL,
		 .pretty_val = NULL,
		 .next_token = NULL,
		 .trans_func = tf_pkcs11_x509file}}},
};

static int pkcs11_get_id_list (char **list) {
	pkcs11h_certificate_id_list_t user_certificates = NULL;
	pkcs11h_certificate_id_list_t current = NULL;
	CK_RV rv = CKR_FUNCTION_FAILED;
	char *s = NULL;
	int rc;

	*list = NULL;

	if (
		(rv = pkcs11h_certificate_enumCertificateIds (
			PKCS11H_ENUM_METHOD_CACHE_EXIST,
			NULL,
			PKCS11H_PROMPT_MASK_ALLOW_ALL,
			NULL,
			&user_certificates
		)) != CKR_OK
	) {
		syslog (LOG_ERR, "PKCS#11: Cannot enumerate certificates %ld-'%s'", rv, pkcs11h_getMessage (rv));
		rc = -EIO;
		goto cleanup;
	}

	for (current = user_certificates; current != NULL; current = current->next) {
		pkcs11h_certificate_t certificate = NULL;
		X509 *x509 = NULL;
		BIO *bio = NULL;
		char dn[1024] = {0};
		char serial[1024] = {0};
		char *ser = NULL;
		char *ssh_key = NULL;
		size_t ser_len = 0;
		int n;

		if (
			(rv = pkcs11h_certificate_serializeCertificateId (
				NULL,
				&ser_len,
				current->certificate_id
			)) != CKR_OK
		) {
			syslog (LOG_ERR, "PKCS#11: Cannot serialize certificate id certificates %ld-'%s'", rv, pkcs11h_getMessage (rv));
			rc = -EIO;
			goto cleanup1;
		}

		if (
			(ser = (char *)malloc (ser_len)) == NULL
		) {
			syslog (LOG_ERR, "PKCS#11: Cannot allocate memory");
			rc = -ENOMEM;
			goto cleanup1;
		}

		if (
			(rv = pkcs11h_certificate_serializeCertificateId (
				ser,
				&ser_len,
				current->certificate_id
			)) != CKR_OK
		) {
			syslog (LOG_ERR, "PKCS#11: Cannot serialize certificate id certificates %ld-'%s'", rv, pkcs11h_getMessage (rv));
			rc = -EIO;
			goto cleanup1;
		}

		if (
			(rv = pkcs11h_certificate_create (
				current->certificate_id,
				NULL,
				PKCS11H_PROMPT_MASK_ALLOW_ALL,
				PKCS11H_PIN_CACHE_INFINITE,
				&certificate
			)) != CKR_OK
		) {
			syslog (LOG_ERR, "PKCS#11: Cannot create certificate %ld-'%s'", rv, pkcs11h_getMessage (rv));
			rc = -EIO;
			goto cleanup1;
		}

		if ((x509 = pkcs11h_openssl_getX509 (certificate)) == NULL) {
			syslog (LOG_ERR, "PKCS#11: Cannot get X509");
			rc = -EIO;
			goto cleanup1;
		}

		X509_NAME_oneline (
			X509_get_subject_name (x509),
			dn,
			sizeof (dn)
		);

		if ((bio = BIO_new (BIO_s_mem ())) == NULL) {
			syslog (LOG_ERR, "PKCS#11: Cannot create BIO");
			rc = -EIO;
			goto cleanup1;
		}

		i2a_ASN1_INTEGER(bio, X509_get_serialNumber (x509));
		n = BIO_read (bio, serial, sizeof (serial)-1);
		if (n<0) {
			serial[0] = '\x0';
		}
		else {
			serial[n] = 0;
		}

		{
			char *t = NULL;

			if (asprintf (&t, "%s%s (%s) [%s]\n", s!=NULL?s:"", dn, serial, ser) == -1) {
				rc = -ENOMEM;
				goto cleanup1;
			}
			if (s != NULL) {
				free (s);
			}
			s = t;
		}

	cleanup1:
		if (x509 != NULL) {
			X509_free (x509);
			x509 = NULL;
		}

		if (bio != NULL) {
			BIO_free_all (bio);
			bio = NULL;
		}

		if (certificate != NULL) {
			pkcs11h_certificate_freeCertificate (certificate);
			certificate = NULL;
		}

		if (ser != NULL) {
			free (ser);
			ser = NULL;
		}
	}

	*list = s;
	s = NULL;
	rc = 0;

cleanup:

	if (user_certificates != NULL) {
		pkcs11h_certificate_freeCertificateIdList (user_certificates);
		user_certificates = NULL;
	}

	if (s != NULL) {
		free (s);
		s = NULL;
	}

	return rc;
}

/**
 * tf_pkcs11_enter
 * @ctx: The current applicable libecryptfs context struct
 * @node: The param_node from which we are transitioning
 * @head: The head of the name/value pair list that is being
 *        constructed as the decision graph is being traversed
 * @foo: Arbitary state information for the current subgraph
 *
 * Each transition from one node in the decision graph to another node
 * can have a function executed on the transition event. A transition
 * into any given subgraph may require certain housekeeping and
 * initialization functions to occur.
 *
 * The decision graph engine forwards along an arbitrary data
 * structure among the nodes of any subgraph. The logic in the
 * subgraph can use that data structure to access and maintain
 * arbitrary status information that is unique to the function of that
 * subgraph.
 */
static int tf_pkcs11_enter(struct ecryptfs_ctx *ctx,
			    struct param_node *param_node,
			    struct val_node **mnt_params, void **foo)
{
	struct ecryptfs_subgraph_ctx *subgraph_ctx;
	int rc;

	if ((subgraph_ctx = malloc(sizeof(struct ecryptfs_subgraph_ctx)))
	    == NULL) {
		rc = -ENOMEM;
		goto out;
	}
	memset(subgraph_ctx, 0, sizeof(struct ecryptfs_subgraph_ctx));
	if ((rc = ecryptfs_find_key_mod(&subgraph_ctx->key_mod, ctx,
					param_node->val))) {
		syslog(LOG_ERR, "%s: Cannot find key_mod for param_node with "
		       "val = [%s]\n", __FUNCTION__, param_node->val);
		goto out;
	}

	if (!strcmp (param_node->mnt_opt_names[0], "key")) {
		if ((rc = pkcs11_get_id_list(&pkcs11_param_nodes[PKCS11_ID_TOK].suggested_val)) != 0) {
			goto out;
		}
	}

	(*foo) = (void *)subgraph_ctx;
out:
	return rc;
}

struct transition_node pkcs11_transition = {
	.val = "pkcs11-helper",
	.pretty_val = "pkcs11-helper module",
	.next_token = &(pkcs11_param_nodes[0]),
	.trans_func = tf_pkcs11_enter
};

static int ecryptfs_pkcs11_get_param_subgraph_trans_node(
	struct transition_node **trans, uint32_t version)
{
	if ((version & ECRYPTFS_VERSIONING_PUBKEY) == 0)
		return -1;
	(*trans) = &pkcs11_transition;
	return 0;
}

int ecryptfs_pkcs11_finalize(void)
{
	if (pkcs11_param_nodes[PKCS11_ID_TOK].suggested_val)
		free(pkcs11_param_nodes[PKCS11_ID_TOK].suggested_val);
	pkcs11h_terminate ();
}

static int ecryptfs_pkcs11_init(char **alias)
{
	char *provider = "/usr/lib/pkcs11/libasepkcs.so";
	PKCS11H_BOOL protected_auth = TRUE;
	int nPINCachePeriod = PKCS11H_PIN_CACHE_INFINITE;
	CK_RV rv = CKR_FUNCTION_FAILED;
	uid_t id;
	struct passwd *pw;
	int rc = 0;

	if (asprintf(alias, "pkcs11-helper") == -1) {
		rc = -ENOMEM;
		syslog(LOG_ERR, "Out of memory\n");
		goto out;
	}
	id = getuid();
	pw = getpwuid(id);
	if (!pw) {
		rc = -EIO;
		goto out;
	}

	if ((rv = pkcs11h_initialize ()) != CKR_OK) {
		syslog(LOG_ERR, "PKCS#11: Cannot initialize %ld-'%s'", rv, pkcs11h_getMessage (rv));
		goto out;
	}

	if ((rv = pkcs11h_setLogHook (pkcs11_log, NULL)) != CKR_OK) {
		syslog(LOG_ERR, "PKCS#11: Cannot set hooks %ld-'%s'", rv, pkcs11h_getMessage (rv));
		goto out;
	}

	pkcs11h_setLogLevel (10);

	if ((rv = pkcs11h_setTokenPromptHook (pkcs11_token_prompt, NULL)) != CKR_OK) {
		syslog(LOG_ERR, "PKCS#11: Cannot set hooks %ld-'%s'", rv, pkcs11h_getMessage (rv));
		goto out;
	}

	if ((rv = pkcs11h_setPINPromptHook (pkcs11_pin_prompt, NULL)) != CKR_OK) {
		syslog(LOG_ERR, "PKCS#11: Cannot set hooks %ld-'%s'", rv, pkcs11h_getMessage (rv));
		goto out;
	}

	if ((rv = pkcs11h_setProtectedAuthentication (protected_auth)) != CKR_OK) {
		syslog(LOG_ERR, "PKCS#11: Cannot set protected authentication mode %ld-'%s'", rv, pkcs11h_getMessage (rv));
		goto out;
	}

	if ((rv = pkcs11h_setPINCachePeriod (nPINCachePeriod)) != CKR_OK) {
		syslog(LOG_ERR, "PKCS#11: Cannot set Pcache period %ld-'%s'", rv, pkcs11h_getMessage (rv));
		goto out;
	}

out:
	if (rc == 0 && rv != CKR_OK) {
		rc = -EIO;
	}

	if (
		(rv = pkcs11h_addProvider (
			provider,
			provider,
			TRUE,
			PKCS11H_PRIVATEMODE_MASK_AUTO,
			PKCS11H_SLOTEVENT_METHOD_AUTO,
			0,
			FALSE
		)) != CKR_OK
	) {
		syslog(LOG_ERR, "PKCS#11: Cannot initialize provider '%s' %ld-'%s'", provider, rv, pkcs11h_getMessage (rv));
	}

	return rc;
}

static int ecryptfs_pkcs11_get_params(struct key_mod_param **params,
				       uint32_t *num_params)
{
	int rc = 0;

	(*params) = NULL;
	(*num_params) = 0;	
out:
	return rc;
}

static struct ecryptfs_key_mod_ops ecryptfs_pkcs11_ops = {
	&ecryptfs_pkcs11_init,
	NULL,
	NULL,
	NULL,
	&ecryptfs_pkcs11_get_param_subgraph_trans_node,
	NULL,
	NULL,
	&ecryptfs_pkcs11_get_key_sig,
	NULL,
	&ecryptfs_pkcs11_encrypt,
	&ecryptfs_pkcs11_decrypt,
	NULL,
	&ecryptfs_pkcs11_finalize
};

struct ecryptfs_key_mod_ops *get_key_mod_ops(void)
{
	return &ecryptfs_pkcs11_ops;
}
